/*----------------------------------------------------------------------
  File    : brreg.c
  Contents: best response regression experiments
  Author  : Christian Borgelt
  History : 2020.03.27 file created
            2020.03.28 bootstrap and gradient functions added
            2020.04.01 Hessian matrix and Newton-Raphson method added
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include "random.h"
#include "normal.h"
#ifndef MAT_WRITE
#define MAT_WRITE
#endif
#include "matrix.h"
#include "params.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#ifndef M_PI                    /* if not defined (no Posix source) */
#define M_PI        3.14159265358979323846  /* \pi */
#endif
#ifndef INFINITY
#define INFINITY    (DBL_MAX+DBL_MAX)
#endif

#define PRGNAME     "brreg"
#define DESCRIPTION "best response regression experiments"
#define VERSION     "version 1.0 (2020.03.28)         " \
                    "(c) 2020   Christian Borgelt"

#define CCHAR       const char  /* abbreviation */

/* --- distribution types --- */
#define DT_CAUCHY    (-2)       /* Cauchy  distribution */
#define DT_LAPLACE   (-1)       /* Laplace distribution */
#define DT_UNIFORM     0        /* uniform distribution */
#define DT_NORMAL      1        /* normal  distribution */

/* --- error codes --- */
/* error codes 0 to -5 defined in attset.h */
#define E_OPTION     (-6)       /* unknown option */
#define E_OPTARG     (-7)       /* missing option argument */
#define E_ARGCNT     (-8)       /* wrong number of arguments */
#define E_OBJFN      (-9)       /* invalid objective function */
#define E_MODE       (-10)      /* invalid operation mode */
#define E_DISTRIB    (-11)      /* invalid distrbution identifier */
#define E_STDDEV     (-12)      /* invalid standard deviation */
#define E_FILE       (-13)      /* error writing data file */
#define E_UNKNOWN    (-14)      /* unknown error */

#define IN_SECS(t)    ((double)(t) /(double)CLOCKS_PER_SEC)
#define SEC_SINCE(t)  IN_SECS(clock()-(t))

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- mode information --- */
  int  code;                    /* code         of mode */
  char *onec;                   /* one character   mode */
  char *abbr;                   /* abbreviation of mode */
  char *name;                   /* name         of mode */
} MODEINFO;                     /* (mode information) */

typedef double OBJFN  (double *a, MATRIX *X, double *y, double *z,
                       double eps, double beta, double gamma);
typedef void   GRADFN (double *a, MATRIX *X, double *y, double *z,
                       double eps, double beta, double gamma,
                       double *g);
typedef void   HESSFN (double *a, MATRIX *X, double *y, double *z,
                       double eps, double beta, double gamma,
                       MATRIX *H);

typedef struct {                /* --- function set for optim. --- */
  OBJFN  *objfn;                /* objective function */
  GRADFN *gradfn;               /* gradient  function */
  HESSFN *hessfn;               /* Hessian matrix function */
} FNSET;

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const MODEINFO modemap[] = { /* table of operation modes */
  { -2, "c",  "c",    "coeffs"    },
  {  0, "x",  "x",    "basic"     },
  /*------------------------------*/
  {  1, "s",  "rbst", "robust"    },
  {  2, "d",  "bbs",  "direct"    },
  /*------------------------------*/
  {  3, "b",  "boot", "bootstrap" },
  {  4, "h",  "hill", "hillclimb" },
  {  5, "g",  "grad", "gradient"  },
  {  6, "r",  "resi", "resilient" },
  {  7, "n",  "newt", "Newton"    },
  { -1, NULL, NULL,   NULL }    /* sentinel */
};

static const MODEINFO objfnmap[] = { /* table of obective functions */
  { -1, "-",  "--",   "--" },
  {  0, "h",  "fh",   "FH" },
  {  1, "l",  "fl",   "FL" },
  {  2, "r",  "fr",   "FR" },
  { -1, NULL, NULL,   NULL }    /* sentinel */
};

static const char *errmsgs[] = {   /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_OBJFN    -9 */  "invalid objective function %s",
  /* E_MODE    -10 */  "invalid operation mode %s",
  /* E_DISTRIB -11 */  "invalid distribution identifier %s",
  /* E_STDDEV  -12 */  "invalid standard deviation %s",
  /* E_FILE    -13 */  "error writing data file",
  /*           -14 */  "unknown error",
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR *prgname;          /* program name for error messages */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/
#ifndef NDEBUG
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static double regval (double *a, double *x, DIMID n)
{                               /* --- compute linear reg. function */
  double r = a[0];              /* result of computation */
  while (--n >= 0) r += a[n+1] *x[n];
  return r;                     /* compute and return function value */
}  /* regval() */

/*--------------------------------------------------------------------*/

static void noise (double *values, DIMID n, int t, double s)
{                               /* --- add noise value of type t */
  DIMID  i;                     /* loop variable */
  int    k;                     /* loop variable */
  double x;                     /* buffer for random value */

  assert(values && t && (s >= 0)); /* check the function arguments */
  if      (t == DT_CAUCHY) {    /* Cauchy distribution */
    for (i = 0; i < n; i++) {   /* traverse the values */
      values[i] += s *2.0 *tan(M_PI *(drand() -0.5));
    } }                         /* add Cauchy distribution sample */
  else if (t == DT_LAPLACE) {   /* Laplace distribution */
    for (i = 0; i < n; i++) {   /* traverse the values */
      x = 2.0 *(drand() -0.5);  /* get sample from uniform [-1,+1] */
      values[i] += s *((x < 0) ? 4.0 : -4.0) *log(1.0 -fabs(x));
    } }                         /* add Laplace distribution sample */
  else if (t == DT_UNIFORM) {   /* uniform distribution */
    for (i = 0; i < n; i++) {   /* traverse the values */
      values[i] += s *2.0 *(drand() -0.5);
    } }                         /* add uniform distribution sample */
  else if (t == DT_NORMAL) {    /* normal distribution */
    for (i = 0; i < n; i++) {   /* traverse the values */
      values[i] += normrand(drand, 0.0, s);
    } }                         /* add normal distribution sample */
  else {                        /* if no known code */
    for (i = 0; i < n; i++) {  /* traverse the values */
      x = normrand(drand, 0.0, 1.0);
      for (k = 0; ++k < t; ) x *= normrand(drand, 0.0, 1.0);
      values[i] += s *x;       /* multiply t samples */
    }                          /* from a normal distribution */
  }                            /* and sum result to the values */
}  /* noise() */

/*--------------------------------------------------------------------*/

static void subdata (double *wgts, DIMID n, DIMID size)
{                               /* --- get random subset as weights */
  DIMID i, k;                   /* loop variable, index */

  memset(wgts, 0, (size_t)n *sizeof(double));
  for (k = 0; k < size; k++) {  /* traverse the sample size */
    i = (int)(n *drand());      /* choose a random index in {0, n-1} */
    if (i >= n) i = n-1;        /* clamp the index */
    if (i <  0) i = 0;          /* to the valid range */
    wgts[i] += 1.0;             /* increment the weight */
  }                             /* of the chosen data point */
}  /* subdata() */

/*----------------------------------------------------------------------
  Regression Function (ordinary least squares, OLS)
----------------------------------------------------------------------*/

static int coeffs (double *res, MATRIX *X, double *y, double *wgts)
{                               /* --- compute regression coeffs. */
  DIMID  m;                     /* number of regressor variables */
  DIMID  i, r, c;               /* loop variables */
  MATRIX *mat;                  /* coefficient matrix */
  double *rhs;                  /* right hand side */
  double *row;                  /* to traverse the data matrix rows */
  double w, t;                  /* data point weight, buffer */
  int    z;                     /* result of function call */

  assert(res && X && y);        /* check the function arguments */
  m   = mat_colcnt(X);          /* get the number of regressors */
  mat = mat_create(m+1, m+1);   /* create a work matrix and */
  if (!mat) return -1;          /* initialize it to zero */
  mat_init(mat, MAT_VECTOR|MAT_ZERO, NULL);
  rhs = mat_vector(mat);        /* get right hand side */
  for (i = mat_rowcnt(X); --i >= 0; ) {
    w = (wgts) ? wgts[i] : 1.0; /* traverse the data points */
    rhs[0] += w *y[i];          /* update right hand side */
    row = mat_row(X, i);        /* get the next data point */
    mat_inc(mat, 0, 0, w);      /* sum the data point weights */
    for (r = 0; r < m; r++) {   /* traverse the matrix rows */
      t = w *row[r];            /* get weighted data values */
      rhs[r+1] += t *y[i];      /* update right hand side */
      mat_inc(mat, 0, r+1, t);  /* update first matrix column */
      for (c = r; c < m; c++)   /* update remaining matrix columns */
        mat_inc(mat, r+1, c+1, t*row[c]);
    }                           /* (build equation system matrix */
  }                             /*  and right hand side) */
  z = mat_chdecx(mat, mat, 0.0, 1e-32, 1e32);
  if (z == 0)                   /* compute Cholesky decomposition */
    mat_chsubst(res, mat, rhs); /* forward/backward substitution */
  mat_delete(mat);              /* delete the coefficient matrix */
  return z;                     /* return decomposition result */
}  /* coeffs() */

/*----------------------------------------------------------------------
  Robust Regression Function (MM estimator)
----------------------------------------------------------------------*/

static double tukey (double e, double k)
{                               /* --- Tukey's bisquare weights */
  if (fabs(e) > k) return 0.0;  /* check against limit, otherwise */
  e /= k; e = 1-e*e; return e*e;/* compute bisquare function */
}  /* tukey() */

/*--------------------------------------------------------------------*/

static int robust (double *a, MATRIX *X, double *y,
                   double thresh, int steps)
{                               /* --- robust regression */
  DIMID  i, k, m, n;            /* loop variables */
  double *wgts;                 /* data point weights */
  double *errs, e;              /* prediction errors */
  double z;                     /* prediction value */
  double s, q;                  /* sum (of squares) of errors */
  double d;                     /* standard deviation of errors */
  double *o;                    /* old regression coefficients */
  int    r;                     /* result of function call */

  n    = mat_rowcnt(X);         /* get the number of data points */
  m    = mat_colcnt(X);         /* and the number of regressors */
  errs = (double*)malloc((size_t)(n+n+m+1) *sizeof(double));
  if (!errs) return -1;         /* allocate work memory and */
  wgts = errs +n; o = wgts +n;  /* organize it into vectors */
  r = coeffs(a, X, y, NULL);    /* compute regression coefficients */
  if (r != 0) { free(errs); return r; }
  // s = q = 0;                    /* initialize sum and sum of squares */
  // for (i = n; --i >= 0; ) {     /* traverse the data points */
  //   z = regval(a, x, m);        /* compute prediction value */
  //   e = z -y[i];                /* compute prediction error and sum */
  //   s += e; q += e*e;           /* errors and squares of errors, */
  // }                             /* then compute standard deviation */
  // d = sqrt((q -s*s/(double)n)/((double)n-1.0));
  for (k = 0; k < steps; k++) { /* compute a certain number of steps */
    s = q = 0;                  /* initialize sum and sum of squares */
    for (i = 0; i < n; i++) {   /* traverse the data points */
      z = regval(a, mat_row(X, i), m); /* compute prediction value, */
      e = errs[i] = fabs(z -y[i]);     /* prediction error and sum */
      s += e; q += e*e;         /* errors and squares of errors, */
    }                           /* then compute standard deviation */
    d = sqrt((q -s*s/(double)n)/((double)n-1.0));
    for (i = 0; i < n; i++)     /* compute data point weights */
      wgts[i] = tukey(errs[i], d);
    memcpy(o, a, (size_t)(m+1) *sizeof(double));
    r = coeffs(a, X, y, wgts);  /* note old regression coefficients */
    if (r != 0) break;          /* and compute new coefficients */
    for (d = 0.0, i = 0; i <= m; i++)
      d += fabs(a[i]-o[i]);     /* sum absolute coefficient changes */
    if (d < thresh) break;      /* check for termination */
  }                             /* (maximum coefficient change) */
  free(errs);                   /* free the work memory */
  return r;                     /* return regression result */
}  /* robust() */

/*----------------------------------------------------------------------
  Best Response Regression Objective Functions
----------------------------------------------------------------------*/

static void H (double *a, double *x, double y, double z, DIMID m,
               double eps, int *l, int *h)
{                               /* --- Heaviside step functions */
  double u = regval(a, x, m);   /* compute prediction and bounds */
  double v = (1.0-eps) *fabs(z-y);
  *l = (u >  y-v) ? 1 : 0;      /* lower Heaviside step function */
  *h = (u >= y+v) ? 1 : 0;      /* upper Heaviside step function */
}  /* H() */

/*--------------------------------------------------------------------*/

static double FH (double *a, MATRIX *X, double *y, double *z,
                  double eps, double beta, double gamma)
{                               /* --- objective function F_H */
  DIMID  i, m;                  /* loop variables */
  int    l, h, s = 0;           /* sum of Heaviside functions */

  for (m = mat_colcnt(X), i = mat_rowcnt(X); --i >= 0; ) {
    H(a, mat_row(X, i), y[i], z[i], m, eps, &l, &h);
    s += l -h;                  /* traverse the data points, */
  }                             /* compute Heaviside functions, */
  return (double)s;             /* and sum their difference */
}  /* FH() */

/*--------------------------------------------------------------------*/

static void L (double *a, double *x, double y, double z, DIMID m,
               double eps, double beta, double *l, double *h)
{                               /* --- logistic functions */
  double u = regval(a, x, m);   /* compute prediction and bounds */
  double v = (1.0-eps) *fabs(z-y);
  double s = -beta *(u -(y-v)); /* compute args. of logistic function */
  double t = -beta *(u -(y+v)); /* and evaluate logistic functions */
  *l = (s < 709.0) ? 1.0 /(1.0 +exp(s)) : 0.0;
  *h = (t < 709.0) ? 1.0 /(1.0 +exp(t)) : 0.0;
}  /* L() */

/*--------------------------------------------------------------------*/

static double FL (double *a, MATRIX *X, double *y, double *z,
                  double eps, double beta, double gamma)
{                               /* --- objective function F_L */
  DIMID  i, m;                  /* loop variables */
  double l, h, s = 0.0;         /* sum of logistic functions */

  for (m = mat_colcnt(X), i = mat_rowcnt(X); --i >= 0; ) {
    L(a, mat_row(X, i), y[i], z[i], m, eps, beta, &l, &h);
    s += l -h;                  /* traverse the data points, */
  }                             /* compute logistic functions, */
  return s;                     /* and sum their difference */
}  /* FL() */

/*--------------------------------------------------------------------*/

static void gradFL (double *a, MATRIX *X, double *y, double *z,
                    double eps, double beta, double gamma, double *g)
{                               /* --- gradient of function F_L */
  DIMID  i, k, m, n;            /* loop variables */
  double l, h, s;               /* values of logistic functions */
  double *x;                    /* current data point */

  n = mat_rowcnt(X);            /* get the number of data points */
  m = mat_colcnt(X);            /* and the number of regressors */
  memset(g, 0, (size_t)(m+1) *sizeof(double));
  for (i = 0; i < n; i++) {     /* traverse the data points */
    x = mat_row(X, i);          /* compute logistic functions */
    L(a, x, y[i], z[i], m, eps, beta, &l, &h);
    g[0] += s = beta *(l*(1-l) -h*(1-h));
    for (k = 0; k < m; k++) g[k+1] += s *x[k];
  }                             /* sum terms of gradient */
}  /* gradFL() */

/*--------------------------------------------------------------------*/

static void HessianFL (double *a, MATRIX *X, double *y, double *z,
                       double eps, double beta, double gamma, MATRIX *H)
{                               /* --- Hessian matrix of function F_L */
  DIMID  i, r, c, m, n;         /* loop variables */
  double l, h, s;               /* values of logistic functions */
  double *x;                    /* current data point */
  double b = beta *beta;        /* common factor of elements */

  n = mat_rowcnt(X);            /* get the number of data points */
  m = mat_colcnt(X);            /* and the number of regressors */
  mat_init(H, MAT_ZERO, NULL);  /* initialize Hessian matrix */
  for (i = 0; i < n; i++) {     /* traverse the data points */
    x = mat_row(X, i);          /* compute logistic functions */
    L(a, x, y[i], z[i], m, eps, beta, &l, &h);
    s = b *(l*(1-l)*(1-2*l) -h*(1-h)*(1-2*h));
    mat_inc(H, 0, 0, s);        /* compute factor for data point */
    for (r = 1; r <= m; r++) {  /* traverse the matrix rows */
      mat_inc(H, r, 0, s *x[r-1]);
      for (c = 1; c <= r; c++)  /* traverse the matrix columns */
        mat_inc(H, r, c, s *x[r-1] *x[c-1]);
    }                           /* only compute lower triangle */
  }                             /* because matrix is symmetric */
  mat_tr2sym(H, H, MAT_LOWER);  /* complete matrix to symmetric */
}  /* HessianFL() */

/*--------------------------------------------------------------------*/

static void R (double *a, double *x, double y, double z, DIMID m,
               double eps, double beta, double gamma,
               double *l, double *h)
{                               /* --- exponential rise functions */
  double u = regval(a, x, m);   /* compute prediction and bounds */
  double v = (1.0-eps) *fabs(z-y);
  double s = u -(y-v);          /* compute args. of exponen. function */
  double t = u -(y+v);          /* evaluate exponential function */
  *l = (s >= 0) ? 1.0 -0.5*exp(-gamma*s) : 0.5*exp(+beta*s);
  *h = (t <= 0) ? 1.0 -0.5*exp(+gamma*t) : 0.5*exp(-beta*t);
}  /* R() */

/*--------------------------------------------------------------------*/

static double FR (double *a, MATRIX *X, double *y, double *z,
                  double eps, double beta, double gamma)
{                               /* --- objective function F_R */
  DIMID  i, m;                  /* loop variables */
  double l, h, s = 0.0;         /* sum of exponential functions */

  for (m = mat_colcnt(X), i = mat_rowcnt(X); --i >= 0; ) {
    R(a, mat_row(X, i), y[i], z[i], m, eps, beta, gamma, &l, &h);
    s += l +h -1;               /* traverse the data points, */
  }                             /* compute exponential functions, */
  return s;                     /* and sum them */
}  /* FR() */

/*--------------------------------------------------------------------*/

static void gradFR (double *a, MATRIX *X, double *y, double *z,
                    double eps, double beta, double gamma, double *g)
{                               /* --- gradient of function F_R */
  DIMID  i, k, m, n;            /* loop variables */
  double l, h, s;               /* values of logistic functions */
  double *x;                    /* current data point */

  n    = mat_rowcnt(X);         /* get the number of data points */
  m    = mat_colcnt(X);         /* and the number of regressors */
  memset(g, 0, (size_t)(m+1) *sizeof(double));
  for (i = 0; i < n; i++) {     /* traverse the data points */
    x = mat_row(X, i);          /* compute logistic functions */
    R(a, x, y[i], z[i], m, eps, beta, gamma, &l, &h);
    g[0] += s = (((l <= 0.5) ? beta*l : gamma*(1-l))
               - ((h <= 0.5) ? beta*h : gamma*(1-h)));
    for (k = 0; k < m; k++) g[k+1] += s *x[k];
  }                             /* sum terms of gradient */
}  /* gradFR() */

/*--------------------------------------------------------------------*/

static void HessianFR (double *a, MATRIX *X, double *y, double *z,
                       double eps, double beta, double gamma, MATRIX *H)
{                               /* --- Hessian matrix of function F_R */
  DIMID  i, r, c, m, n;         /* loop variables */
  double l, h, s;               /* values of logistic functions */
  double *x;                    /* current data point */
  double b = beta  *beta;       /* common factor of elements */
  double g = gamma *gamma;      /* common factor of elements */

  n = mat_rowcnt(X);            /* get the number of data points */
  m = mat_colcnt(X);            /* and the number of regressors */
  mat_init(H, MAT_ZERO, NULL);  /* initialize Hessian matrix */
  for (i = 0; i < n; i++) {     /* traverse the data points */
    x = mat_row(X, i);          /* compute logistic functions */
    R(a, x, y[i], z[i], m, eps, beta, gamma, &l, &h);
    s = (((l <= 0.5) ? b*l : g*(l-1))
       + ((h <= 0.5) ? b*h : g*(h-1)));
    mat_inc(H, 0, 0, s);        /* compute factor for data point */
    for (r = 1; r <= m; r++) {  /* traverse the matrix rows */
      mat_inc(H, r, 0, s *x[r-1]);
      for (c = 1; c <= r; c++)  /* traverse the matrix columns */
        mat_inc(H, r, c, s *x[r-1] *x[c-1]);
    }                           /* only compute lower triangle */
  }                             /* because matrix is symmetric */
  mat_tr2sym(H, H, MAT_LOWER);  /* complete matrix to symmetric */
}  /* HessianFR() */

/*--------------------------------------------------------------------*/

static FNSET fnsets[] = {       /* function sets for optimization */
  { FH, NULL,   NULL },         /* with objective function, */
  { FL, gradFL, HessianFL },    /* gradient function, and */
  { FR, gradFR, HessianFR } };  /* Hessian matrix function */

/*----------------------------------------------------------------------
  Best Response Regression Optimization Functions
----------------------------------------------------------------------*/

static int optBSdirect (double *a, MATRIX *X, double *y, double *z,
                        double eps, int trials)
{                               /* --- bootstrap optimization */
  DIMID  i, m, n;               /* loop variables */
  double *eref;                 /* reference errors */
  double *wgts;                 /* weights for bootstrap sample */
  double *c;                    /* current regression coefficients */
  DIMID  w, w_best;             /* (best) number of wins */
  double s, s_best;             /* (best) sum of absolute errors */
  double p, e;                  /* buffers for prediction and error */
  int    r;                     /* result of function call */

  n    = mat_rowcnt(X);         /* get the number of data points */
  m    = mat_colcnt(X);         /* and the number of regressors */
  wgts = (double*)malloc((size_t)(n+n+m+1) *sizeof(double));
  if (!wgts) return -1;         /* allocate work memory and */
  c = wgts +n; eref = c +m;     /* organize it into vectors */
  r = coeffs(a, X, y, NULL);    /* ordinary regression as default */
  if (r != 0) { free(wgts); return r; }
  w = 0; s = 0.0;               /* initialize win counter */
  for (i = 0; i < n; i++) {     /* traverse the data points */
    eref[i] = fabs(y[i] -z[i]); /* compute the reference errors */
    p = regval(a, mat_row(X, i), m);
    e = fabs(y[i] -p);          /* compute prediction and its error */
    if (e < eref[i] *(1.0 -eps)) {
      w += 1; s += e; }         /* count won data points and */
  }                             /* sum error of these data points */
  w_best = w; s_best = s;       /* current values are best seen */
  while (--trials >= 0) {       /* execute multiple trials */
    subdata(wgts, n, n);        /* create bootstrap sample as weights */
    r = coeffs(c, X, y, wgts);  /* compute regression coefficients */
    if (r != 0) break;          /* abort on failure */
    w = 0; s = 0.0;             /* initialize win counter */
    for (i = 0; i < n; i++) {   /* traverse the data points */
      p = regval(c, mat_row(X, i), m);
      e = fabs(y[i] -p);        /* compute prediction and its error */
      if (e < eref[i] *(1.0 -eps)) {
        w += 1; s += e; }       /* count won data points and */
    }                           /* sum error of these data points */
    if ((w > w_best) || ((w == w_best) && (s < s_best))) {
      w_best = w; s_best = s;   /* if current coefficients are better */
      memcpy(a, c, (size_t)(m+1) *sizeof(double));
    }                           /* update best regression coeffs. */
  }
  free(wgts);                   /* free work memory */
  return r;                     /* return regression result */
}  /* optBSdirect() */

/*--------------------------------------------------------------------*/

static int optBootstrap (double *a, FNSET *fnset,
                         MATRIX *X, double *y, double *z,
                         double eps, double beta, double gamma,
                         int trials)
{                               /* --- bootstrap optimization on F_L */
  DIMID  n, m;                  /* number of data points/regressors */
  double *wgts;                 /* weights for bootstrap sample */
  double *c;                    /* current regression coefficients */
  double best, q;               /* (best) value of function F_L */
  int    r;                     /* result of function call */

  n    = mat_rowcnt(X);         /* get the number of data points */
  m    = mat_colcnt(X);         /* and the number of regressors */
  wgts = (double*)malloc((size_t)(n+m+1) *sizeof(double));
  if (!wgts) return -1;         /* allocate work memory and */
  c = wgts +n;                  /* organize it into vectors */
  r = coeffs(a, X, y, NULL);    /* ordinary regression as default */
  if (r != 0) { free(wgts); return r; }
  best = fnset->objfn(a, X, y, z, eps, beta, gamma);
                                /* evaluate regression coeffs. */
  while (--trials >= 0) {       /* execute multiple trials */
    subdata(wgts, n, n);        /* create bootstrap sample as weights */
    r = coeffs(c, X, y, wgts);  /* compute regression coefficients */
    if (r != 0) break;          /* abort on failure */
    q = fnset->objfn(c, X, y, z, eps, beta, gamma);
    if (q > best) {             /* evaluate regression coefficients */
      best = q; memcpy(a, c, (size_t)(m+1) *sizeof(double)); }
  }                             /* update best regression coeffs. */
  free(wgts);                   /* free work memory */
  return r;                     /* return regression result */
}  /* optBootstrap() */

/*--------------------------------------------------------------------*/

static int optGrad (double *a, FNSET *fnset,
                    MATRIX *X, double *y, double *z,
                    double eps, double beta, double gamma,
                    int trials, double eta, double thresh, int steps)
{                               /* --- standard gradient descent */
  DIMID  i, k, n, m;            /* loop variables */
  double *wgts;                 /* weights for bootstrap sample */
  double *c;                    /* current regression coefficients */
  double *g;                    /* gradient of regression coeffs. */
  double best, q;               /* (best) value of function F_L */
  double s, t;                  /* temporary buffers */
  int    r;                     /* result of function call */

  n    = mat_rowcnt(X);         /* get the number of data points */
  m    = mat_colcnt(X);         /* and the number of regressors */
  wgts = (double*)malloc((size_t)(n+m+m+2) *sizeof(double));
  if (!wgts) return -1;         /* allocate work memory and */
  c = wgts +n; g = c +m+1;      /* organize it into vectors */
  r = coeffs(a, X, y, NULL);    /* compute regression coefficients */
  if (r != 0) { free(wgts); return r; }
  memcpy(c, a, (size_t)(m+1) *sizeof(double));
  best = fnset->objfn(a, X, y, z, eps, beta, gamma);
  while (1) {                   /* evaluate regression coeffs. */
    for (k = 0; k < steps; k++){/* do a maximum number of updates */
      fnset->gradfn(c, X, y, z, eps, beta, gamma, g);
      for (s = 0, i = 0; i <= m; i++) {    /* compute gradient and */
        c[i] += t = eta *g[i]; s += t*t; } /* update coefficients */
      if (sqrt(s) < thresh) break;  /* check for small gradient */
      if ((k >= 8) && (k % 4 == 0)  /* check objective function */
      &&  (fnset->objfn(c, X, y, z, eps, beta, gamma) < best)) break;
    }                           /* abort if improvement is small */
    q = fnset->objfn(c, X, y, z, eps, beta, gamma);
    if (q > best) {             /* evaluate regression coefficients */
      best = q; memcpy(a, c, (size_t)(m+1) *sizeof(double)); }
    if (--trials < 0) break;    /* execute multiple trials */
    subdata(wgts, n, n);        /* create bootstrap sample as weights */
    r = coeffs(c, X, y, wgts);  /* compute regression coefficients */
    if (r != 0) break;          /* and abort on failure */
  }
  free(wgts);                   /* free working memory */
  return r;                     /* return regression result */
}  /* optGrad() */

/*--------------------------------------------------------------------*/

static int optResilient (double *a, FNSET *fnset,
                         MATRIX *X, double *y, double *z,
                         double eps, double beta, double gamma,
                         int trials, double eta, double thresh,
                         double shrink, double growth,
                         double minchg, double maxchg, int steps)
{                               /* --- resilient gradient descent */
  DIMID  i, k, n, m;            /* loop variables */
  double *wgts;                 /* weights for bootstrap sample */
  double *c;                    /* current regression coefficients */
  double *g, *p;                /* gradient of regression coeffs. */
  double *e;                    /* step widths/learning rates */
  double best, q;               /* (best) value of function F_L */
  double s, t;                  /* temporary buffers */
  int    r;                     /* result of function call */

  n    = mat_rowcnt(X);         /* get the number of data points */
  m    = mat_colcnt(X);         /* and the number of regressors */
  wgts = (double*)malloc((size_t)(n+4*m+4) *sizeof(double));
  if (!wgts) return -1;         /* allocate work memory and */
  c = wgts +n; g = c +m+1;      /* organize it into vectors */
  p = g +m+1;  e = p +m+1;      /* (gradient, previous, change) */
  r = coeffs(a, X, y, NULL);    /* compute regression coefficients */
  if (r != 0) { free(wgts); return r; }
  memcpy(c, a, (size_t)(m+1) *sizeof(double));
  best = fnset->objfn(a, X, y, z, eps, beta, gamma);
  while (1) {                   /* evaluate regression coeffs. */
    for (i = 0; i <= m; i++) {  /* init. step width/learning rates */
      e[i] = eta; p[i] = 0.0; } /* and clear previous gradient */
    s = 0;                      /* initialize total change */
    for (k = 0; k < steps; k++){/* do a maximum number of updates */
      fnset->gradfn(c, X, y, z, eps, beta, gamma, g);
      for (i = 0; i <= m; i++){ /* traverse the parameters */
        if      (g[i] > 0) t =  p[i];
        else if (g[i] < 0) t = -p[i];
        else               t =  0;
        if      (t > 0) {       /* if gradients have the same sign, */
          e[i] *= growth;       /* increase the step width */
          if (e[i] > maxchg) e[i] = maxchg;
          p[i] = g[i]; }        /* note the current gradient */
        else if (t < 0) {       /* if gradients have opposite signs, */
          e[i] *= shrink;       /* decrease the step width */
          if (e[i] < minchg) e[i] = minchg;
          p[i] = 0.0; }         /* suppress a change in the next step */
        else                    /* if one gradient is zero, */
          p[i] = g[i];          /* only note the current gradient */
          /* SuperSAB */
          /* a[i] += e[i] *g[i]; */  /* update the regression coeffs. */
          /* Resilient */
        if      (g[i] > 0) c[i] += e[i];
        else if (g[i] < 0) c[i] -= e[i];
        s += e[i]*e[i];         /* add steps to regression coeffs. */
      }                         /* and sum squared changes */
      if (sqrt(s) < thresh) break;
      if ((k >= 8) && (k % 4 == 0)  /* check objective function */
      &&  (fnset->objfn(c, X, y, z, eps, beta, gamma) < best)) break;
    }
    q = fnset->objfn(c, X, y, z, eps, beta, gamma);
    if (q > best) {             /* evaluate regression coefficients */
      best = q; memcpy(a, c, (size_t)(m+1) *sizeof(double)); }
    if (--trials < 0) break;    /* execute multiple trials */
    subdata(wgts, n, n);        /* create bootstrap sample as weights */
    r = coeffs(c, X, y, wgts);  /* compute regression coefficients */
    if (r != 0) break;          /* and abort on failure */
  }
  free(wgts);                   /* free working memory */
  return r;                     /* return regression result */
}  /* optResilient() */

/*--------------------------------------------------------------------*/

static int optNewton (double *a, FNSET *fnset,
                      MATRIX *X, double *y, double *z,
                      double eps, double beta, double gamma,
                      int trials, double thresh, int steps)
{                               /* --- Newton-Raphson on gradient */
  DIMID  i, k, n, m;            /* loop variables */
  double *wgts;                 /* weights for bootstrap sample */
  double *c;                    /* current regression coefficients */
  double *g;                    /* gradient of regression coeffs. */
  MATRIX *H;                    /* Hessian matrix */
  double best, q;               /* (best) value of function F_L */
  double s;                     /* temporary buffers */
  int    r;                     /* result of function call */

  n    = mat_rowcnt(X);         /* get the number of data points */
  m    = mat_colcnt(X);         /* and the number of regressors */
  wgts = (double*)malloc((size_t)(n+m+m+2) *sizeof(double));
  if (!wgts) return -1;         /* allocate work memory and */
  c = wgts +n; g = c +m+1;      /* organize it into vectors */
  H = mat_create(m+1, m+1);     /* create a matrix for the Hessian */
  if (!H) { free(wgts); return -1; }
  r = coeffs(a, X, y, NULL);    /* compute regression coefficients */
  if (r != 0) { free(wgts); return r; }
  memcpy(c, a, (size_t)(m+1) *sizeof(double));
  best = fnset->objfn(a, X, y, z, eps, beta, gamma);
  while (1) {                   /* evaluate regression coeffs. */
    s = 0;                      /* initialize total change */
    for (k = 0; k < steps; k++){/* do a maximum number of updates */
      fnset->hessfn(c, X, y, z, eps, beta, gamma, H);
      r = mat_gjinv(H, H, 0, NULL); /* compute inverse of Hessian */
      if (r != 0) { r = 0; break; } /* matrix and gradient vector */
      fnset->gradfn(c, X, y, z, eps, beta, gamma, g);
      mat_mulmv(g, H, g);       /* compute change vector and */
      for (i = 0; i <= m; i++){ /* add it to the regression coeffs. */
        c[i] -= g[i]; s += g[i]*g[i]; }
      if (sqrt(s) < thresh) break;
      if ((k >= 8) && (k % 4 == 0)  /* check objective function */
      &&  (fnset->objfn(c, X, y, z, eps, beta, gamma) < best)) break;
    }
    q = fnset->objfn(c, X, y, z, eps, beta, gamma);
    if (q > best) {             /* evaluate regression coefficients */
      best = q; memcpy(a, c, (size_t)(m+1) *sizeof(double)); }
    if (--trials < 0) break;    /* execute multiple trials */
    subdata(wgts, n, n);        /* create bootstrap sample as weights */
    r = coeffs(c, X, y, wgts);  /* compute regression coefficients */
    if (r != 0) break;          /* and abort on failure */
  }
  mat_delete(H);                /* delete Hessian matrix */
  free(wgts);                   /* free working memory */
  return r;                     /* return regression result */
}  /* optNewton() */

/*----------------------------------------------------------------------
  Basic Tests
----------------------------------------------------------------------*/

static int ctest (void)
{                               /* --- basic test of coeffs function */
  DIMID  i, m = 1, n;           /* loop variables */
  double x[] = { 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0 };
  double y[] = { 1.0, 3.0, 2.0, 3.0, 4.0, 3.0, 5.0, 6.0 };
  MATRIX *X;                    /* data matrix */
  double a[2] = {0.0,0.0}, z;   /* regression coefficients/value */

  n = sizeof(y)/sizeof(y[0]);   /* get number of data points */
  X = mat_create(n, 1);         /* create a data matrix */
  if (!X) return -1;            /* with a single column */
  mat_init(X, MAT_FULL, x);     /* and initialize it */
  if (coeffs(a, X, y, NULL) != 0)
    return -2;                  /* compute regression coefficients */
  printf("% 8.6f % 8.6f\n", a[0], a[1]);   /* and print the result */
  for (i = 0; i < mat_rowcnt(X); i++) {
    z = regval(a, mat_row(X, i), m);
    printf("% 8.6f % 8.6f % 8.6f\n", mat_get(X, i, 0), y[i], z);
  }                             /* compute & print prediction values */
  return 0;                     /* return 'ok' */
}  /* ctest() */

/*--------------------------------------------------------------------*/

static void showres (double *a, MATRIX *X, double *y, double *z,
                     double eps, double beta, double gamma, clock_t t)
{                               /* --- show optimization result */
  printf("% 9.6f % 9.6f (%.3fs)\n", a[0], a[1], IN_SECS(t));
  printf("%2g %9.6f %9.6f\n", FH(a, X, y, z, eps, beta, gamma),
                              FL(a, X, y, z, eps, beta, gamma),
                              FR(a, X, y, z, eps, beta, gamma));
}  /* showres() */

/*--------------------------------------------------------------------*/

static int basic (double eps, double beta, double gamma,
                  int trials, double eta, double thresh, int steps,
                  double shrink, double growth,
                  double minchg, double maxchg)
{                               /* --- basic test of functionality */
  DIMID   i, m = 1, n;          /* loop variables */
  double  x[] = { 0.0, 0.1, 0.2, 0.3, 0.4,
                  0.5, 0.6, 0.7, 0.8, 0.9, 5.0 };
  double  y[] = { 1.54901800, 0.07267671,-0.03379786,-1.77214300,
                 -1.56598000,-2.89350300,-4.90599200,-4.82752800,
                 -6.44050900,-7.06924500, 5.00000000 };
  double  z[] = { 0.0, 0.0, 0.0, 0.0, 0.0,
                  0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
  double  w[] = { 1.0, 1.0, 1.0, 1.0, 1.0,
                  1.0, 1.0, 1.0, 1.0, 1.0, 0.0 };
  MATRIX  *X;                   /* data matrix */
  double  a[2];                 /* regression coefficients */
  int     r;                    /* result of function call */
  clock_t t;                    /* for time measurements */

  n = sizeof(y)/sizeof(y[0]);   /* get number of data points */
  X = mat_create(n, 1);         /* create a data matrix */
  if (!X) return -1;            /* with a single column */
  mat_init(X, MAT_FULL, x);     /* initialize data matrix */

  printf("\nordinary regression:\n");
  t = clock();                  /* start timer */
  r = coeffs(a, X, y, NULL);    /* compute regression coefficients */
  if (r != 0) { mat_delete(X); return r; }
  t = clock() -t;               /* and compute prediction values */
  for (i = 0; i < mat_rowcnt(X); i++)
    z[i] = regval(a, mat_row(X, i), m);
  showres(a, X, y, z, eps, beta, gamma, t);

  printf("\nordinary regression without last point:\n");
  t = clock();                  /* start timer */
  r = coeffs(a, X, y, w);       /* compute regression coefficients */
  if (r != 0) { mat_delete(X); return r; }
  t = clock() -t;               /* and evaluate the result */
  showres(a, X, y, z, eps, beta, gamma, t);

  printf("\nrobust regression:\n");
  t = clock();                  /* start timer */
  r = robust(a, X, y, thresh, steps);
  if (r != 0) { mat_delete(X); return r; }
  t = clock() -t;               /* and evaluate the result */
  showres(a, X, y, z, eps, beta, gamma, t);

  printf("\nbootstrap sampling (direct):\n");
  t = clock();                  /* start timer */
  r = optBSdirect(a, X, y, z, eps, 10*trials);
  if (r != 0) { mat_delete(X); return r; }
  t = clock() -t;               /* and evaluate the result */
  showres(a, X, y, z, eps, beta, gamma, t);

  printf("\nbootstrap sampling F_H:\n");
  t = clock();                  /* start timer */
  r = optBootstrap(a, fnsets+0, X, y, z, eps, beta, gamma, 10*trials);
  if (r != 0) { mat_delete(X); return r; }
  t = clock() -t;               /* and evaluate the result */
  showres(a, X, y, z, eps, beta, gamma, t);

  printf("\nbootstrap sampling F_L:\n");
  t = clock();                  /* start timer */
  r = optBootstrap(a, fnsets+1, X, y, z, eps, beta, gamma, 10*trials);
  if (r != 0) { mat_delete(X); return r; }
  t = clock() -t;               /* and evaluate the result */
  showres(a, X, y, z, eps, beta, gamma, t);

  printf("\nbootstrap sampling F_R:\n");
  t = clock();                  /* start timer */
  r = optBootstrap(a, fnsets+2, X, y, z, eps, beta, gamma, 10*trials);
  if (r != 0) { mat_delete(X); return r; }
  t = clock() -t;               /* and evaluate the result */
  showres(a, X, y, z, eps, beta, gamma, t);

  printf("\nstandard gradient descent F_L:\n");
  t = clock();                  /* start timer */
  r = optGrad(a, fnsets+1, X, y, z, eps, beta, gamma,
              trials, eta, thresh, steps);
  if (r != 0) { mat_delete(X); return r; }
  t = clock() -t;               /* and evaluate the result */
  showres(a, X, y, z, eps, beta, gamma, t);

  printf("\nresilient gradient descent F_L:\n");
  t = clock();                  /* start timer */
  r = optResilient(a, fnsets+1, X, y, z, eps, beta, gamma,
                   trials, eta, thresh,
                   shrink, growth, minchg, maxchg, steps);
  if (r != 0) { mat_delete(X); return r; }
  t = clock() -t;               /* and evaluate the result */
  showres(a, X, y, z, eps, beta, gamma, t);

  printf("\nNewton_Raphson on gradient of F_L:\n");
  t = clock();                  /* start timer */
  r = optNewton(a, fnsets+1, X, y, z, eps, beta, gamma,
                trials, thresh, steps);
  if (r != 0) { mat_delete(X); return r; }
  t = clock() -t;               /* and evaluate the result */
  showres(a, X, y, z, eps, beta, gamma, t);

  printf("\nstandard gradient descent F_R:\n");
  t = clock();                  /* start timer */
  r = optGrad(a, fnsets+2, X, y, z, eps, beta, gamma,
              trials, eta, thresh, steps);
  if (r != 0) { mat_delete(X); return r; }
  t = clock() -t;               /* and evaluate the result */
  showres(a, X, y, z, eps, beta, gamma, t);

  printf("\nresilient gradient descent F_R:\n");
  t = clock();                  /* start timer */
  r = optResilient(a, fnsets+2, X, y, z, eps, beta, gamma,
                   trials, eta, thresh,
                   shrink, growth, minchg, maxchg, steps);
  if (r != 0) { mat_delete(X); return r; }
  t = clock() -t;               /* and evaluate the result */
  showres(a, X, y, z, eps, beta, gamma, t);

  printf("\nNewton_Raphson on gradient of F_R:\n");
  t = clock();                  /* start timer */
  r = optNewton(a, fnsets+2, X, y, z, eps, beta, gamma,
                trials, thresh, steps);
  if (r != 0) { mat_delete(X); return r; }
  t = clock() -t;               /* and evaluate the result */
  showres(a, X, y, z, eps, beta, gamma, t);

  mat_delete(X);                /* delete the data matrix */
  return 0;                     /* return 'ok' */
}  /* basic() */

/*----------------------------------------------------------------------
  Extended Tests
----------------------------------------------------------------------*/

static void gendata (double *a, MATRIX *X, double *y,
                     int xdist, double xsdev)
{                               /* --- generate data set */
  DIMID  i, n;                  /* loop variables */
  double t;                     /* buffer */

  n = mat_rowcnt(X);            /* get number of data points */
  if (xdist == DT_NORMAL) {     /* if normal distribution requested */
    for (i = n; --i >= 0; )     /* sample x-values from normal dist. */
      mat_set(X, i, 0, normrand(drand, 0, xsdev)); }
  else {                        /* if uniform distribution requested */
    t = 2.0 *sqrt(3.0) *xsdev;  /* get spread of values */
    for (i = n; --i >= 0; )     /* sample x-values from uniform dist. */
      mat_set(X, i, 0, t *(drand() -0.5));
  }                             /* initialize the data matrix */
  for (i = n; --i >= 0; )       /* compute the output values */
    y[i] = a[0] +a[1] *mat_get(X, i, 0);
}  /* gendata() */

/*--------------------------------------------------------------------*/

static int extra (int mode, double *a, DIMID size,
                  int xdist, double xsdev, int edist, double esdev,
                  int objfn, double eps, double beta, double gamma,
                  int trials, double eta, double thresh, int steps,
                  double shrink, double growth,
                  double minchg, double maxchg,
                  int runs, CCHAR *ddir)
{                               /* --- best response regression exps. */
  DIMID  i, k;                  /* loop variables */
  MATRIX *X;                    /* data matrix */
  double *y, *z;                /* output values (true and predicted) */
  FNSET  *fnset;                /* fn. set for objective function */
  double *ca, *cb;              /* coefficients */
  double *d;                    /* residuals for ordinary regression */
  double u, v;                  /* prediction values */
  double ea, eb;                /* sums of squared errors */
  int    wn;                    /* number of won points */
  int    wtrn, wtst, wsse;      /* number of runs won */
  int    r;                     /* result of function call */
  char   *fname = NULL;         /* name of data file */
  FILE   *out   = NULL;         /* data file */
  char   *xname = NULL;         /* name of x-value distribution */
  char   *ename = NULL;         /* name of error   distribution */
  char   buf[16];               /* buffer for number output */

  y = (double*)malloc((size_t)(5*size+4) *sizeof(double));
  if (!y) return -1;            /* allocate work memory */
  X = mat_create(size, 1);      /* and the data matrix */
  if (!X) { free(y); return -1; }
  if (ddir) {                   /* if to write data files */
    fname = (char*)malloc(strlen(ddir) +128);
    if (!fname) { mat_delete(X); free(y); return -1; }
    xname = (xdist == DT_NORMAL)  ? "n" : "u";
    ename = (edist <= DT_CAUCHY)  ? "c"
          : (edist <= DT_LAPLACE) ? "l"
          : (edist <= DT_UNIFORM) ? "u"
          : (edist <= DT_NORMAL)  ? "n" : "";
    if (edist > DT_NORMAL)      /* power of normal distribution */
      sprintf(ename = buf, "%d", edist-DT_NORMAL-1);
  }                             /* create a buffer for file names */
  z  = y +size; d  = z  +size;  /* organize work memory */
  ca = d +size; cb = ca +2;
  fnset = fnsets +objfn;        /* get objective function set */
  wtrn = wtst = wsse = 0;       /* initialize the win counters */
  for (k = 0; k < runs; k++) {  /* execute runs (traverse data sets) */
    gendata(a, X, y, xdist, xsdev);  /* generate a data set */
    noise(y, size, edist, esdev);    /* and add some noise */
    if (ddir) {                 /* if to save the created data */
      sprintf(fname, "%strn_%g_%g_n%"DIMID_FMT"_x%s%g_e%s%g_r%d.txt",
              ddir, a[0], a[1], size, xname, xsdev, ename, esdev, k);
      out = fopen(fname, "w");  /* create file name, open data file */
      if (!out) { free(fname); mat_delete(X); free(y); return -2; }
      for (i = 0; i < size; i++)/* save the data */
        fprintf(out, "%g\t%g\n", mat_get(X, i, 0), y[i]);
      if (fclose(out) != 0) { /* close the data file */
        free(fname); mat_delete(X); free(y); return -2; }
    }
    r = coeffs(ca, X, y, NULL); /* compute regression coefficients */
    if (r != 0) { mat_delete(X); free(y); return r; }
    for (wn = 0, i = 0; i < size; i++) {
      z[i] = ca[0] +ca[1] *mat_get(X, i, 0);
      d[i] = fabs(z[i]-y[i]);   /* compute OLS regression */
    }                           /* and its residuals */
    if      (mode == 1)         /* robust regression */
      r = robust      (cb, X, y, thresh, steps);
    else if (mode == 2)         /* direct bootstrap */
      r = optBSdirect (cb, X, y, z, eps, trials);
    else if (mode == 3)         /* objective function bootstrap */
      r = optBootstrap(cb, fnset, X, y, z,
                       eps, beta, gamma, trials);
    /* else if (mode == 4)         ** hill climbing **
      r = optClimb    (cb, fnset, X, y, z, eps, beta, gamma,
                       trials, eta, steps); */
    else if (mode == 5)         /* gradient ascent */
      r = optGrad     (cb, fnset, X, y, z, eps, beta, gamma,
                       trials, eta, thresh, steps);
    else if (mode == 6)         /* resilient gradient ascent */
      r = optResilient(cb, fnset, X, y, z, eps, beta, gamma,
                       trials, eta, thresh,
                       shrink, growth, minchg, maxchg, steps);
    else if (mode == 7)         /* Newton-Raphson on gradient */
      r = optNewton   (cb, fnset, X, y, z, eps, beta, gamma,
                       trials, thresh, steps);
    for (wn = 0, i = 0; i < size; i++) {
      u = fabs(cb[0] +cb[1] *mat_get(X, i, 0) -y[i]);
      if (u < d[i]) wn++;       /* compute regression predictions */
    }                           /* and compare to OLS residuals */
    if (wn+wn > size) wtrn++;   /* count won training runs */

    gendata(a, X, y, xdist, xsdev);  /* generate a data set */
    noise(y, size, edist, esdev);    /* and add some noise */
    if (ddir) {                 /* if to save the created data */
      sprintf(fname, "%stst_%g_%g_n%"DIMID_FMT"_x%s%g_e%s%g_r%d.txt",
              ddir, a[0], a[1], size, xname, xsdev, ename, esdev, k);
      out = fopen(fname, "w");  /* create file name, open data file */
      if (!out) { free(fname); mat_delete(X); free(y); return -2; }
      for (i = 0; i < size; i++)/* save the data */
        fprintf(out, "%g\t%g\n", mat_get(X, i, 0), y[i]);
      if (fclose(out) != 0) { /* close the data file */
        free(fname); mat_delete(X); free(y); return -2; }
    }
    for (ea = eb = 0, wn = 0, i = 0; i < size; i++) {
      u = fabs(ca[0] +ca[1] *mat_get(X, i, 0) -y[i]);
      v = fabs(cb[0] +cb[1] *mat_get(X, i, 0) -y[i]);
      ea += u*u; eb += v*v;     /* compute regression predictions, */
      if (v < u) wn++;          /* their residuals and their sses */
    }                           /* and compare the residuals */
    if (wn+wn > size) wtst++;   /* count won test runs */
    if (eb    < ea)   wsse++;   /* count won test runs (sse) */
    //u = (double)(k+1);          /* compute and print win rates */
    //printf("%3d %5.3f %5.3f %5.3f\n", runs, wtrn/u, wtst/u, wsse/u);
  }
  u = (double)runs;             /* compute and print win rates */
  printf("%3d %5.3f %5.3f %5.3f\n", runs, wtrn/u, wtst/u, wsse/u);

  if (fname) free(fname);       /* delete the data file name */
  mat_delete(X);                /* delete the data matrix */
  free(y);                      /* and the output vectors */
  return 0;                     /* return 'ok' */
}  /* extra() */

/*----------------------------------------------------------------------
  Main Program
----------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counter */
  char    *s, *t;               /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *ddir  = NULL;        /* data directory (optional) */
  CCHAR   *mode  = "basic";     /* operation mode */
  int     mcode  = 0;           /* operation mode code */
  double  a[]    = {0.0,1.0};   /* true linear function parameters */
  DIMID   size   = 100;         /* number of data points */
  CCHAR   *xdist = "uniform";   /* distribution of indep. variable */
  int     xcode  = DT_UNIFORM;  /* code of distribution */
  double  xsdev  = 2.0;         /* standard deviation of indep. var. */
  CCHAR   *edist = "normal";    /* distribution of error/noise */
  int     ecode  = DT_UNIFORM;  /* code of distribution */
  double  esdev  = 1.0;         /* standard deviation of error/noise */
  CCHAR   *objfn = "FH";        /* objective function */
  int     fncode = 0;           /* objective function code */
  double  eps    = 0.001;       /* minimum improvement percentage */
  double  beta   = 1.0;         /* steepness parameter 1 (beta) */
  double  gamma  = 32.0;        /* steepness parameter 2 (gamma) */
  int     trials = 20;          /* number of trials/bootstrap samples */
  double  eta    = 0.001;       /* learning rate eta */
  double  thresh = 1e-6;        /* threshold for termination */
  int     steps  = 100;         /* number of update steps */
  double  shrink = 0.5;         /* shrink and */
  double  growth = 1.2;         /* growth factor */
  double  minchg = 1e-6;        /* minimum and */
  double  maxchg = 1.0;         /* maximum change */
  int     runs   = 1000;        /* number of runs / repetitions */
  long    seed;                 /* seed for random numbers */
  double  x;                    /* exchange buffer */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument is given */
    printf("usage: %s [options] [mode]\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-a#      true linear function coefficients      "
                    "(default: %g,%g)\n", a[0], a[1]);
    printf("-n#      number of data points                  "
                    "(default: %d)\n",    size);
    printf("-x#      distribution of independent variable   "
                    "(default: %s)\n",    xdist);
    printf("-e#      noise to be added and its scale        "
                    "(default: %s)\n",    edist);
    printf("-o#      objective function                     "
                    "(default: %s)\n",    objfn);
    printf("-i#      minimum improvement percentage epsilon "
                    "(default: %g)\n",    eps);
    printf("-b#      steepness parameter beta               "
                    "(default: %g)\n",    beta);
    printf("-g#      steepness parameter gamma              "
                    "(default: %g)\n",    gamma);
    printf("-t#      number of trials / bootstrap samples   "
                    "(default: %d)\n",    trials);
    printf("-l#      learning rate eta                      "
                    "(default: %g)\n",    eta);
    printf("-T#      threshold for termination              "
                    "(default: %g)\n",    thresh);
    printf("-s#      number of update steps                 "
                    "(default: %d)\n",    steps);
    printf("-f#      growth and shrink factors              "
                    "(default: %g:%g)\n", growth, shrink);
    printf("-z#      minimum and maximum change             "
                    "(default: %g:%g)\n", minchg, maxchg);
    printf("-r#      number of runs / repetitions           "
                    "(default: %d)\n",    runs);
    printf("-S#      seed value for random number generator "
                    "(default: time)\n");
    printf("-d#      path for writing data files            "
                    "(default: do not write)\n");
    printf("mode     parameter optimization method          "
                    "(default: basic test)\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  seed = (long)time(NULL);      /* get a default seed value */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case 'a': getdblvec(s, &s, 2, a);              break;
          case 'n': size   = (DIMID)strtol(s, &s, 0);    break;
          case 'x': optarg = &xdist;                     break;
          case 'e': optarg = &edist;                     break;
          case 'o': optarg = &objfn;                     break;
          case 'i': eps    =        strtod(s, &s);       break;
          case 'b': beta   =        strtod(s, &s);       break;
          case 'g': gamma  =        strtod(s, &s);       break;
          case 't': trials = (int)  strtol(s, &s, 0);    break;
          case 'l': eta    =        strtod(s, &s);       break;
          case 'T': thresh =        strtod(s, &s);       break;
          case 's': steps  = (int)  strtol(s, &s, 0);    break;
          case 'f': getdbls(s, &s, 2, &shrink, &growth); break;
          case 'z': getdbls(s, &s, 2, &minchg, &maxchg); break;
          case 'r': runs   = (int)  strtol(s, &s, 0);    break;
          case 'S': seed   =        strtol(s, &s, 0);    break;
          case 'd': optarg = &ddir;                      break;
          default : error(E_OPTION, *--s);               break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: mode = s;        break;
        default: error(E_ARGCNT); break;
      }                         /* note file names */
    }
  }
  if (optarg) error(E_OPTARG);  /* check option argument */
  rseed((unsigned)seed);        /* init. the random number generator */
  if (growth < shrink) { x = growth; growth = shrink; shrink = x; }
  if (maxchg < minchg) { x = maxchg; maxchg = minchg; maxchg = x; }
  for (i = 0; modemap[i].name; i++)
    if ((strcmp(modemap[i].onec, mode) == 0)
    ||  (strcmp(modemap[i].abbr, mode) == 0)
    ||  (strcmp(modemap[i].name, mode) == 0)) break;
  if (modemap[i].name)          /* search for the optimization method */
    mcode = modemap[i].code;    /* and if found, get the method code */
  else {                        /* if no method name or abbreviation */
    mcode = (int)strtol(mode, &s, 0);
    if (*s || (s == mode)) error(E_MODE, mode);
  }                             /* get method code directly */
  for (i = 0; objfnmap[i].name; i++)
    if ((strcmp(objfnmap[i].onec, objfn) == 0)
    ||  (strcmp(objfnmap[i].abbr, objfn) == 0)
    ||  (strcmp(objfnmap[i].name, objfn) == 0)) break;
  if (objfnmap[i].name)         /* search for the optimization method */
    fncode = objfnmap[i].code;  /* and if found, abort with failur */
  else error(E_OBJFN, objfn);   /* get method code directly */
  s = strchr(xdist, ':');       /* check for two part argument */
  if (s) {                      /* if standard deviation is present */
    *s++ = 0; xsdev = strtod(s, &t);
    if (*t || (t == s)) error(E_STDDEV, s);
  }                             /* extract standard deviation */
  if      ((strcmp(xdist, "n")       == 0)
  ||       (strcmp(xdist, "N")       == 0)
  ||       (strcmp(xdist, "normal")  == 0)
  ||       (strcmp(xdist, "Normal")  == 0)
  ||       (strcmp(xdist, "g")       == 0)
  ||       (strcmp(xdist, "G")       == 0)
  ||       (strcmp(xdist, "gauss")   == 0)
  ||       (strcmp(xdist, "Gauss")   == 0)) xcode = DT_NORMAL;
  else if ((strcmp(xdist, "u")       == 0)
  ||       (strcmp(xdist, "U")       == 0)
  ||       (strcmp(xdist, "uniform") == 0)
  ||       (strcmp(xdist, "Uniform") == 0)) xcode = DT_UNIFORM;
  else error(E_DISTRIB, xdist); /* get x-distribution code */
  s = strchr(edist, ':');       /* check for two-part argument */
  if (s) {                      /* if standard deviation is present */
    *s++ = 0; esdev = strtod(s, &t);
    if (*t || (t == s)) error(E_STDDEV, s);
  }                             /* extract standard deviation */
  if      ((strcmp(edist, "c")       == 0)
  ||       (strcmp(edist, "C")       == 0)
  ||       (strcmp(edist, "cauchy")  == 0)
  ||       (strcmp(edist, "Cauchy")  == 0)) ecode = DT_CAUCHY;
  else if ((strcmp(edist, "l")       == 0)
  ||       (strcmp(edist, "L")       == 0)
  ||       (strcmp(edist, "laplace") == 0)
  ||       (strcmp(edist, "Laplace") == 0)) ecode = DT_LAPLACE;
  else if ((strcmp(edist, "u")       == 0)
  ||       (strcmp(edist, "U")       == 0)
  ||       (strcmp(edist, "uniform") == 0)
  ||       (strcmp(edist, "Uniform") == 0)) ecode = DT_UNIFORM;
  else if ((strcmp(edist, "n")       == 0)
  ||       (strcmp(edist, "N")       == 0)
  ||       (strcmp(edist, "normal")  == 0)
  ||       (strcmp(edist, "Normal")  == 0)
  ||       (strcmp(edist, "g")       == 0)
  ||       (strcmp(edist, "G")       == 0)
  ||       (strcmp(edist, "gauss")   == 0)
  ||       (strcmp(edist, "Gauss")   == 0)) ecode = DT_NORMAL;
  else {                        /* get error distribution type */
    ecode = (int)strtol(edist, &s, 0);
    if (*t || (t == s)) error(E_DISTRIB, edist);
  }                             /* check for a number */
  fputc('\n', stderr);          /* terminate the startup message */

  switch (mcode) {              /* evaluate the operation mode */
    case -1: k = ctest();                                      break;
    case  0: k = basic(eps, beta, gamma, trials, eta, thresh,
                       steps, shrink, growth, minchg, maxchg); break;
    default: k = extra(mcode, a, size,
                       xcode, xsdev, ecode, esdev,
                       fncode, eps, beta, gamma,
                       trials, eta, thresh, steps,
                       shrink, growth, minchg, maxchg,
                       runs,   ddir);                          break;
  }                             /* call corresponding function */
  if (k == -1) error(E_NOMEM);  /* check for an error */
  if (k == -2) error(E_FILE);
  if (k <   0) error(E_UNKNOWN);

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
